@@ -67,7 +67,8 @@ enum
 static gchar *dump_value  (JsonGenerator *generator,
                            gint           level,
                            const gchar   *name,
-                           JsonNode      *node);
+                           JsonNode      *node,
+                           gsize         *length);
 static gchar *dump_array  (JsonGenerator *generator,
                            gint           level,
                            const gchar   *name,
@@ -265,7 +266,8 @@ static gchar *
 dump_value (JsonGenerator *generator,
             gint           level,
             const gchar   *name,
-            JsonNode      *node)
+            JsonNode      *node,
+            gsize         *length)
 {
   JsonGeneratorPrivate *priv = generator->priv;
   gboolean pretty = priv->pretty;
@@ -307,10 +309,11 @@ dump_value (JsonGenerator *generator,
 
     case G_TYPE_DOUBLE:
       {
-        gchar buf[65];
+        gchar buf[G_ASCII_DTOSTR_BUF_SIZE];
 
-        g_ascii_formatd (buf, 65, "%g", g_value_get_double (&value));
-        g_string_append (buffer, buf);
+        g_string_append (buffer,
+                         g_ascii_dtostr (buf, sizeof (buf),
+                                         g_value_get_double (&value)));
       }
       break;
 
@@ -325,6 +328,9 @@ dump_value (JsonGenerator *generator,
 
   g_value_unset (&value);
 
+  if (length)
+    *length = buffer->len;
+
   return g_string_free (buffer, FALSE);
 }
 
@@ -379,7 +385,7 @@ dump_array (JsonGenerator *generator,
           break;
 
         case JSON_NODE_VALUE:
-          value = dump_value (generator, sub_level, NULL, cur);
+          value = dump_value (generator, sub_level, NULL, cur, NULL);
           g_string_append (buffer, value);
           g_free (value);
           break;
@@ -474,7 +480,7 @@ dump_object (JsonGenerator *generator,
           break;
 
         case JSON_NODE_VALUE:
-          value = dump_value (generator, sub_level, member_name, cur);
+          value = dump_value (generator, sub_level, member_name, cur, NULL);
           g_string_append (buffer, value);
           g_free (value);
           break;
@@ -581,7 +587,7 @@ json_generator_to_data (JsonGenerator *generator,
       break;
 
     case JSON_NODE_VALUE:
-      retval = NULL;
+      retval = dump_value (generator, 0, NULL, root, length);
       break;
     }
 
